# copyright ############################### #
# This file is part of the Xwakes Package.  #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from pathlib import Path

_pkg_root = Path(__file__).parent.absolute()

